#ifndef GUARD_playlist_h
#define GUARD_playlist_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playlist.h
/// @brief Class for Playlist header file.

#include <string>
#include <cstring>
#include <queue>

#include <gtkmm/treeview.h>
#include <gtkmm/treemodel.h>
#include <gtkmm/liststore.h>
#include <gtkmm/menu.h>

#include "misc.h"

class MainWnd;
class Tag;
class TagFactory;

/// @brief Class for Playlist.
/// @todo Reduce relationship to MainWnd.
class Playlist : public Gtk::TreeView
{
    public:
	Playlist(MainWnd& mainWnd, const std::string& listId, const unsigned int& titleWidth = 470, const bool& loadlist = true);
	virtual ~Playlist();

	void AddFile(const std::string& filePath, const bool& addListFile);
	void AddFiles(const std::vector<std::string>& files, const bool& addListFile);
	void AddDirectory(const std::string& directoryPath, const bool& recursive);

	/// @brief Remove all items.
	void RemoveAll() 
	{   
	    OnMenuRemoveAll(); 
	}

	/// @brief Get file path.
	std::string Get()
	{
	    return GetCurrentPositionFilename();
	}
	std::string GetNext();
	std::string GetPrev();
	std::string GetRandom();
	void ForceUpdateToppage();
	void ForceUpdatePage();

	void ShowCurrentFileInfoDlg() const;

    private:
	class ModelColumns : public Gtk::TreeModel::ColumnRecord
	{
	    public:
		ModelColumns()
		{
		    add(tagParsed_);
		    add(id_);
		    add(filename_);
		    add(title_);
		    add(time_);
		    add(timeSeconds_);
		}

		Gtk::TreeModelColumn<bool> tagParsed_;
		Gtk::TreeModelColumn<unsigned int> id_;
		Gtk::TreeModelColumn<std::string> filename_;
		Gtk::TreeModelColumn<Glib::ustring> title_;
		Gtk::TreeModelColumn<Glib::ustring> time_;
		Gtk::TreeModelColumn<unsigned int> timeSeconds_;
	};

	class EachRow 
	{
	    public:
		EachRow(const Gtk::TreeModel::iterator& iter, const Glib::ustring& title, const Glib::ustring& time, const unsigned int& timeSeconds)
		    : iter_(iter), title_(title), time_(time), timeSeconds_(timeSeconds)
		    {}

		Gtk::TreeModel::iterator iter_;
		Glib::ustring title_;
		Glib::ustring time_;
		unsigned int timeSeconds_;
	};

	const std::string fileProtocol_;		///< String to determine file protocol, "file://"
	const std::string listId_;			///< Identification string of list.

	ModelColumns columns_;

	Gtk::Menu menuPopup_;
	Gtk::TreeModel::iterator pos_;

	MainWnd& mainWnd_;

	Glib::Thread* tagParsingThread_;
	Glib::Mutex tagParsingMutex_;
	Glib::Dispatcher signalRowToFill_;
	std::queue<EachRow> rowQueue_;
	void FillRowFromQueue(); 

	unsigned int count_;				///< Number of added items.
	unsigned int tagProcessCount_;			///< Number of items reading tag at one thread.

	bool inTagParsing_;				///< Tag parsing thread is running.
	bool stopTagParsing_;				///< To stop tag parsing.

	Glib::RefPtr<Gtk::ListStore> treeModel_;

	const std::auto_ptr<TagFactory> tagFactory_;

	virtual bool on_button_press_event(GdkEventButton* ev);
	virtual void on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column);
	virtual void on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj);
	virtual void on_size_allocate(Gtk::Allocation& allocation);

	/// @brief Filename compare function to sort list.
	int FilenameCompare(const Gtk::TreeModel::iterator& a, const Gtk::TreeModel::iterator& b)
	{
	    std::string strA = (*a)[columns_.filename_];
	    std::string strB = (*b)[columns_.filename_];

	    std::string aPath = Glib::path_get_dirname(strA);
	    std::string bPath = Glib::path_get_dirname(strB);

	    // If files in same directory, sort them
	    // or sort for directory.	
	    int rtn = std::strncmp(aPath.c_str(), bPath.c_str(), aPath.size());
	    if(rtn == 0)
		return std::strncmp(strA.c_str(), strB.c_str(), strA.size());
	    else	
		return rtn;
	}

	/// @brief Move cursor to coresspond iterator and scroll to that.
	void MoveCursorAndScroll(const Gtk::TreeModel::iterator& iter)
	{
	    Gtk::TreeModel::Path path;
	    if(iter) {
		path = Gtk::TreeModel::Path(iter);
	    } else {
		path = Gtk::TreeModel::Path("0");
	    }
	    set_cursor(path);
	    scroll_to_row(path, 0.5);
	}

	std::string GetCurrentPositionFilename(); 
	void ParseTag(Gtk::TreeModel::iterator& iter, const bool& thread);
	void GetInfoFromTag(Gtk::TreeModel::iterator& iter);

	void ShowFileInfoDlg(const std::string& filename) const;

	void OnDragDataReceived(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y, const Gtk::SelectionData& selectionData, guint info, guint time);

	void OnMenuRemove();
	void OnMenuRemoveAll();
	void OnMenuSortFilename();
	void OnMenuInfo();

	void OnVScrolled();

	void LoadLastList();

	void LoadList(const std::string& filename);
	void SaveList();

	void StopThread();

	void CreatePopupMenuList();
	void RegisterTagFactory();

	bool IsListEmpty()
	{
	    if(count_ == 0)
		return true;

	    if(!treeModel_->children().begin())
		return true;

	    return false;
	}

	bool IsPositionLastRow() const
	{
	    return pos_ == --(treeModel_->children().end());
	}

	bool IsPositionFirstRow() const
	{
	    return pos_ == treeModel_->children().begin();
	}

	std::string GetNextRowFilename()
	{
	    ++pos_;

	    return GetCurrentPositionFilename();
	}

	std::string GetPrevRowFilename()
	{
	    pos_--;

	    return GetCurrentPositionFilename();
	}

	std::string GetLastRowFilename()
	{
	    pos_ = --(treeModel_->children().end());

	    return GetCurrentPositionFilename();
	}

	/// @brief Calculate row number per page.
	/// @todo Remove magic constant.
	void CalcTagProcessCount(const unsigned int& height)
	{
	    tagProcessCount_ = height / 25 + 2;
	}

	void FillRow(Gtk::TreeModel::iterator& iter, const Glib::ustring& title, const Glib::ustring& totalTime, const unsigned long& timeSeconds)
	{
	    (*iter)[columns_.title_] = title;
	    (*iter)[columns_.time_] = totalTime;
	    (*iter)[columns_.timeSeconds_] = timeSeconds;

	    if(timeSeconds != 1)
		(*iter)[columns_.tagParsed_] = true;
	}

	Gtk::TreeModel::iterator GetTopPositionIterator()
	{
	    Gtk::TreePath eachPath;
	    Gtk::TreeViewColumn* eachColumn = NULL;
	    int cellX = 0;
	    int cellY = 0;

	    get_path_at_pos(0, 0, eachPath, eachColumn, cellX, cellY);

	    return treeModel_->get_iter(eachPath);
	}

	void AddDragedFiles(const std::string& dragedString);

	bool IsLineComment(const std::string& line)
	{
	    if(line[0] == '#')
		return true;

	    return false;
	}

	void TrimLastSpaceCharacter(std::string& dest)
	{
	    const std::string::reverse_iterator space = find_if(dest.rbegin(), dest.rend(), std::not1(IsCharSpace()));
	    if(space != dest.rend() && space != dest.rbegin()) {
		dest = std::string(dest.begin(), space.base());

	    }
	}
};

#endif

