// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file misc.cc
/// @brief Miscellaneous functions.

#include <fstream>

#include "../include/misc.h"

/// @brief Convert string to UTF8 interface function.
///
/// Try to convert 'orig' from machine's charset to UTF8. 
/// If convert is failed, retry to convert 'fail' to UTF8.
///
/// If machine's charset is UTF8 and 'orig' is not UTF8, try to convert
/// 'orig' from 'assume' charset to UTF8.
///
/// @param orig Original string to convert.
/// @param fail Retry string when convert is failed.
/// @param assume Assumption charset.
/// @return Converted string.
Glib::ustring StringToUTF8(const Glib::ustring& orig, const std::string& fail, const std::string& assume)
{
    static std::string currentCharSet;
    static bool isCharSetUTF8;

    if(orig.validate()) 
	return orig;
    else {
	if(currentCharSet.empty())
	    isCharSetUTF8 = Glib::get_charset(currentCharSet);

	if(isCharSetUTF8)
	    return StringToUTF8WithFail(orig, fail, assume);
	else
	    return StringToUTF8WithFail(orig, fail, currentCharSet);
    } 
}


/// @brief Convert string to UTF8 with fail string.
/// @param orig Original string to convert.
/// @param fail Retry string when convert is failed.
/// @param charset From charset.
/// @return Converted string.
/// @sa Glib::ustring StringToUTF8
inline Glib::ustring StringToUTF8WithFail(const Glib::ustring& orig, const std::string& fail, const std::string& charset)
{
    if(fail.empty()) {

	return StringToUTF8WithoutFail(orig, charset);

    } else {

	try {

	    return Glib::convert_with_fallback(orig, "UTF-8", charset);

	} catch (Glib::ConvertError& err) {

	    return StringToUTF8WithoutFail(fail, charset);

	}

    }
}

/// @brief Convert string to UTF8 without fail string.
/// @param orig Original string to convert.
/// @param charset From charset.
/// @return Converted string.
/// @sa Glib::ustring StringToUTF8
inline Glib::ustring StringToUTF8WithoutFail(const Glib::ustring& orig, const std::string& charset)
{
    try {

	return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch (Glib::ConvertError& err) {

	return Glib::ustring();

    }
}

/// @brief Split each directory element and push it to container.
/// @param fullpath Path to split.
/// @param dest Destination container.
void SplitEachDirToVector(const std::string& fullpath, std::vector<std::string>& dest)
{
    typedef std::string::const_iterator iter;

    iter begin = fullpath.begin();
    const iter end = fullpath.end();
    iter temp;
    while(begin != end)
    {
	begin = find_if(begin, end, std::not1(IsCharSlash()));
	temp = find_if(begin, end, IsCharSlash());
	if(begin != end)
	    dest.push_back(std::string(begin, temp));
	begin = temp;		
    }

}

/// @brief Get filenames in directory and push to container.
/// @param path Directory to get filename.
/// @param dest Destination container.
/// @param recursive Get filenames recursively.
void GetFilenameFromDirToVector(const std::string& path, std::vector<std::string>& dest, bool recursive)
{
    Glib::Dir dir(path);

    std::string fullpath;
    // Very ugly repeat codes...
    if(recursive) {
	for(Glib::Dir::const_iterator eachFile = dir.begin(); eachFile != dir.end(); ++eachFile)
	{
	    fullpath = path;
	    fullpath += '/';
	    fullpath += *eachFile;

	    if(IsDirectory(fullpath)) {
		// Repeat again or recursive call?
		if(IsExecutable(fullpath) && IsReadable(fullpath)) {
		    Glib::Dir dir2(fullpath);
		    
		    std::string fullpath2;
		    for(Glib::Dir::const_iterator eachFile2 = dir2.begin(); eachFile2 != dir2.end(); ++eachFile2)
		    {
			fullpath2 = fullpath;
			fullpath2 += '/';
			fullpath2 += *eachFile2;

			if(IsDirectory(fullpath2)) {
			    if(IsExecutable(fullpath2) && IsReadable(fullpath2))
				// If have more depth, call again anyway.
				GetFilenameFromDirToVector(fullpath2, dest, true);
			} else
			    dest.push_back(fullpath2);
		    }
		}
	    } else
		dest.push_back(fullpath);
	}
    } else {
	for(Glib::Dir::const_iterator eachFile = dir.begin(); eachFile != dir.end(); ++eachFile)
	{
	    fullpath = path;
	    fullpath += '/';
	    fullpath += *eachFile;

	    if(!IsDirectory(fullpath))
		dest.push_back(fullpath);
	}
    }

}

/// @brief Read config file and put to map.
/// @param filename Config filename.
/// @param configMap Destination map.
/// @param convertLowerCase Convert key and value to lower case.
/// @param valueBreakAtSpace Read value to space character. 
/// @return Success of open config file.
bool ConfigfileToMap(const std::string& filename, std::map<std::string, std::string>& configMap, const bool& convertLowerCase, const bool& valueBreakAtSpace)
{
    std::ifstream prefFile(filename.c_str());

    if(prefFile) {
	typedef std::string::iterator iter;
	std::string eachLine;

	iter i, j, k;
	std::string key;

	while(std::getline(prefFile, eachLine)) {
	    // Pass comment line, but there is no comment probably unless 
	    // edited by user. Because i don't write comment line to config file. :)
	    if(eachLine[0] == '#')
		continue;

	    // Convert string to lower case
	    if(convertLowerCase) 
		for_each(eachLine.begin(), eachLine.end(), ToLower());

	    i = eachLine.begin();
	    // Eleminate front space character 
	    i = find_if(i, eachLine.end(), std::not1(IsCharSpace()));

	    // Find '=' character
	    j = find_if(i, eachLine.end(), IsCharEqual());
	    if(i != eachLine.end() && i != j) {
		// Eleminate space character between i to j and that is key to insert. 
		k = find_if(i, j, IsCharSpace());
		key = std::string(i, k);

		// Find value to insert.
		i = ++j;
		i = find_if(i, eachLine.end(), std::not1(IsCharSpace()));

		if(valueBreakAtSpace)
		    j = find_if(i, eachLine.end(), IsCharSpace());
		else
		    j = eachLine.end();

		if(i != eachLine.end()) {
		    configMap[key] = std::string(i, j);
		}
	    }
	}

	return true;

    } else
	return false;
}

