// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_aac.cc
/// @brief Class for mpeg 4 audio decoding using FAAD2 source file.

#include <cstring>
#include <string>

#include <vector>

#include "../include/player_aac.h"
#include "../include/mainmanager.h"
#include "../include/scopeguard.h"

FAAD::AACPlayer::AACPlayer(MainManager& mainManager)
: decodeState_(STOP),
    inverseSampleRate_(0.0),
    currentSeconds_(0),
    seekSeconds_(0),
    sampleRate_(0),
    totalSampleCount_(0),
    frameSize_(1024),
    mp4File_(false),
    channels_(0),
    inFile_(NULL),
    sampleBuffer_(NULL),
    mainManager_(mainManager)
{
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalTimerUpdate_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayTimeUpdate));
}

FAAD::AACPlayer::~AACPlayer()
{
    Clean();
}

/// @brief Initialize player.
///
/// Open destination file and determine type.
bool FAAD::AACPlayer::Init(const std::string& filename)
{
    inFile_ = std::fopen(filename.c_str(), "rb");
    if(inFile_ == NULL)
	return false;

    mp4File_ = IsMP4File(inFile_);

    return true;
}

void FAAD::AACPlayer::Start()
{
    if(inFile_ != NULL) {
	decodeState_ = PLAY;

	if(mp4File_) {

	    MP4Decode();

	} else {

	    AACDecode();

	}

	if(decodeState_ == PLAY)
	    mainManager_.AudioClose(true);
	else
	    mainManager_.AudioClose(false);	

    }

    signalFinish_();
}

void FAAD::AACPlayer::Stop()
{
    decodeState_ = STOP;
}

void FAAD::AACPlayer::Seek(const unsigned long& seconds)
{
    decodeState_ = SEEK;
    seekSeconds_ = seconds;
}

const long int& FAAD::AACPlayer::CurrentTime() const
{
    return currentSeconds_;
}

void FAAD::AACPlayer::Clean()
{
    if(inFile_ != NULL)
	std::fclose(inFile_);
}

/// @brief Decode MP4 type file.
/// @todo Consider if have more than 2 channels.
/// @todo Need real testing. I have few MP4 encoded file, and works fine with them. But will work with others?
void FAAD::AACPlayer::MP4Decode()
{
    const std::auto_ptr<mp4ff_callback_t> mp4Cb = SetMP4ffCallback(inFile_);

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    SetDecoderConfig(decoder);

    mp4ff_t* const inFile = mp4ff_open_read(mp4Cb.get());
    if(inFile == NULL)
	return;
    ScopeGuard closeMP4ffFile = MakeGuard(mp4ff_close, inFile);

    int track = GetAACTrack(inFile);
    if(track < 0)
	return;

    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;
    mp4ff_get_decoder_config(inFile, track, &buffer, &bufferSize);

    if(faacDecInit2(decoder, buffer, bufferSize, &sampleRate_, &channels_) < 0)
	return;

    mp4AudioSpecificConfig mp4ASC;
    if(buffer != NULL) {
	if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC) >= 0) {
	    if(mp4ASC.frameLengthFlag == 1)
		frameSize_ = 960;
	    if(mp4ASC.sbr_present_flag == 1)
		frameSize_ *= 2;	
	}

	free(buffer);
	buffer = NULL;
	bufferSize = 0;
    }

    const long numSamples = mp4ff_num_samples(inFile, track);
    int rc = 0;
    long seconds = 0;
    faacDecFrameInfo frameInfo;
    mainManager_.AudioOpenDevice(16, static_cast<unsigned int>(channels_), sampleRate_, mainManager_.audioFormatNativeEndian_);
    inverseSampleRate_ = 1.0 / sampleRate_;
    for(long sampleId = 0; sampleId < numSamples; ++sampleId) {
	if(decodeState_ == STOP)
	    break;

	if(decodeState_ == SEEK) {

	    sampleId = ConvertSecondToCount(seekSeconds_, mp4ASC.samplingFrequency, frameSize_);

	    if(sampleId < 0)
		sampleId = 0;

	    if(sampleId > numSamples)
		break;

	    totalSampleCount_ = sampleId * 2000;		
	    seconds = seekSeconds_;

	    decodeState_ = PLAY;

	}

	rc = mp4ff_read_sample(inFile, track, sampleId, &buffer, &bufferSize);
	if(rc == 0)
	    return;

	sampleBuffer_ = faacDecDecode(decoder, &frameInfo, buffer, bufferSize);

	if(buffer != NULL) {
	    free(buffer);
	    buffer = NULL;
	}

	totalSampleCount_ += frameInfo.samples;
	seconds = ConvertCountToSecond(totalSampleCount_, frameSize_, inverseSampleRate_);
	if(currentSeconds_ != seconds) {
	    currentSeconds_ = seconds;
	    signalTimerUpdate_();
	}

	if(frameInfo.error == 0 && frameInfo.samples > 0) 
	    mainManager_.AudioPlay(reinterpret_cast<double*>(sampleBuffer_), frameInfo.samples>>1);

    }
}

/// @brief Determine type.
bool FAAD::AACPlayer::IsMP4File(FILE* const inFile)
{
    unsigned char header[8];
    std::fread(header, 1, 8, inFile);
    ScopeGuard rewindFile = MakeGuard(std::rewind, inFile);
    if(header[4] == 'f' && header[5] == 't' && header[6] == 'y' && header[7] == 'p')
	return true;

    return false;
}

int FAAD::AACPlayer::GetAACTrack(mp4ff_t* const inFile)
{
    const int totalTracks = mp4ff_total_tracks(inFile);
    int rc = 0;
    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;
    mp4AudioSpecificConfig mp4ASC;

    for(int i = 0; i < totalTracks; ++i) {
	mp4ff_get_decoder_config(inFile, i, &buffer, &bufferSize);

	if(buffer != NULL) {
	    rc = AudioSpecificConfig(buffer, bufferSize, &mp4ASC);
	    free(buffer);
	    buffer = NULL;

	    if(rc >= 0)
		return i;
	}
    }

    return -1;
}

/// @brief Decode AAC type file.
/// @todo Consider if have more than 2 channels.
/// @todo Need real testing. Implement seek function.
void FAAD::AACPlayer::AACDecode()
{
    aac_buffer aacBuffer;
    aacBuffer.infile = inFile_;

    const int maxChannels = 2;
    aacBuffer.buffer = static_cast<unsigned char*>(std::malloc(FAAD_MIN_STREAMSIZE * maxChannels));
    if(!aacBuffer.buffer)
	return;
    std::memset(aacBuffer.buffer, 0, FAAD_MIN_STREAMSIZE * maxChannels);

    int bread = std::fread(aacBuffer.buffer, 1, FAAD_MIN_STREAMSIZE * maxChannels, aacBuffer.infile);
    aacBuffer.bytes_into_buffer = bread;
    aacBuffer.bytes_consumed = 0;
    aacBuffer.file_offset = 0;

    if(bread != FAAD_MIN_STREAMSIZE * maxChannels)
	aacBuffer.at_eof = 1;

    int tagSize = 0;
    if(memcmp(aacBuffer.buffer, "ID3", 3) != 0) {
	tagSize = (aacBuffer.buffer[6] << 21) | (aacBuffer.buffer[7] << 14) |
	    (aacBuffer.buffer[8] << 7) | (aacBuffer.buffer[9] << 0);

	tagSize += 10;
	AACAdvanceBuffer(&aacBuffer, tagSize);
	AACFillBuffer(&aacBuffer);
    }

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    SetDecoderConfig(decoder);

    if(aacBuffer.buffer[0] == 0xFF && (aacBuffer.buffer[1] & 0xF6) == 0xF0) {
	int frameLength;
	for(int frames = 0;; ++frames) {
	    AACFillBuffer(&aacBuffer);
	    if(aacBuffer.bytes_into_buffer > 7) {
		if(!(aacBuffer.buffer[0] == 0xFF && (aacBuffer.buffer[1] & 0xF6) == 0XF0))
		    break;

		frameLength = ((static_cast<unsigned int>(aacBuffer.buffer[3]) & 0x3) << 11) |
		    (static_cast<unsigned int>(aacBuffer.buffer[4]) << 3) |
		    aacBuffer.buffer[5] >> 5;
		if(frameLength > aacBuffer.bytes_into_buffer)
		    break;

		AACAdvanceBuffer(&aacBuffer, frameLength);

	    } else
		break;
	}

	std::fseek(aacBuffer.infile, tagSize, SEEK_SET);

	bread = std::fread(aacBuffer.buffer, 1, FAAD_MIN_STREAMSIZE * maxChannels, aacBuffer.infile);
	if(bread != FAAD_MIN_STREAMSIZE * maxChannels)
	    aacBuffer.at_eof = 1;
	else
	    aacBuffer.at_eof = 0;

	aacBuffer.bytes_into_buffer = bread;
	aacBuffer.bytes_consumed = 0;
	aacBuffer.file_offset = tagSize;        

    }

    AACFillBuffer(&aacBuffer);
    bread = faacDecInit(decoder, aacBuffer.buffer, aacBuffer.bytes_into_buffer,
	    &sampleRate_, &channels_);
    if(bread < 0) 
	return;

    AACAdvanceBuffer(&aacBuffer, bread);
    AACFillBuffer(&aacBuffer);

    long seconds = 0;
    faacDecFrameInfo frameInfo;
    mainManager_.AudioOpenDevice(16, static_cast<unsigned int>(channels_), sampleRate_, mainManager_.audioFormatNativeEndian_);
    inverseSampleRate_ = 1.0 / sampleRate_;
    do {
	if(decodeState_ == STOP)
	    break;

	sampleBuffer_ = faacDecDecode(decoder, &frameInfo, aacBuffer.buffer,
		aacBuffer.bytes_into_buffer);

	AACAdvanceBuffer(&aacBuffer, frameInfo.bytesconsumed);

	totalSampleCount_ += frameInfo.samples;
	seconds = ConvertCountToSecond(totalSampleCount_, frameSize_, inverseSampleRate_);
	if(currentSeconds_ != seconds) {
	    currentSeconds_ = seconds;
	    signalTimerUpdate_();
	}

	if(frameInfo.error == 0 && frameInfo.samples > 0) 
	    mainManager_.AudioPlay(reinterpret_cast<double*>(sampleBuffer_), frameInfo.samples>>1);

	AACFillBuffer(&aacBuffer);
	if(aacBuffer.bytes_into_buffer == 0)
	    sampleBuffer_ = NULL;

    } while(sampleBuffer_ != NULL);

}

void FAAD::AACPlayer::AACFillBuffer(aac_buffer* const aacBuffer)
{
    if(aacBuffer->bytes_consumed > 0) {
	if(aacBuffer->bytes_into_buffer) {
	    memmove(static_cast<void*>(aacBuffer->buffer), 
		    static_cast<void*>(aacBuffer->buffer + aacBuffer->bytes_consumed),
		    aacBuffer->bytes_into_buffer * sizeof(unsigned char));

	}

	if(!aacBuffer->at_eof) {
	    int bread = std::fread(static_cast<void*>(aacBuffer->buffer + aacBuffer->bytes_into_buffer),
		    1, aacBuffer->bytes_consumed, aacBuffer->infile);

	    if(bread != aacBuffer->bytes_consumed)
		aacBuffer->at_eof = 1;

	    aacBuffer->bytes_into_buffer += bread;	

	}

	aacBuffer->bytes_consumed = 0;

	if(aacBuffer->bytes_into_buffer > 3) {
	    if(memcmp(aacBuffer->buffer, "TAG", 3) == 0)
		aacBuffer->bytes_into_buffer = 0;
	}

	if(aacBuffer->bytes_into_buffer > 11) {
	    if(memcmp(aacBuffer->buffer, "LYRICSBEGIN", 11) == 0)
		aacBuffer->bytes_into_buffer = 0;
	}

	if(aacBuffer->bytes_into_buffer > 8) {
	    if(memcmp(aacBuffer->buffer, "APETAGEX", 8) == 0)
		aacBuffer->bytes_into_buffer = 0;
	}
    }
}

void FAAD::AACPlayer::AACAdvanceBuffer(aac_buffer* const aacBuffer, const int& bytes)
{
    aacBuffer->file_offset += bytes;
    aacBuffer->bytes_consumed = bytes;
    aacBuffer->bytes_into_buffer -= bytes;
}

/// @brief Set mp4_callback_t.
std::auto_ptr<mp4ff_callback_t> FAAD::AACPlayer::SetMP4ffCallback(FILE* inFile)
{
    std::auto_ptr<mp4ff_callback_t> mp4Cb(new mp4ff_callback_t);
    mp4Cb->read = MP4ReadCallback;
    mp4Cb->seek = MP4SeekCallback;
    mp4Cb->user_data = inFile;

    return mp4Cb;
}

/// @brief Set decoder configuration.
/// @param decoder Decoder.
/// @param defObjectType Set object type.
/// @param outputFormat Library output formats. 
/// @param downMatrix Change multichannels to less one.
/// @param useOldADTSFormat Use old ADTS format.
/// @param dontUpSampleImplicitSBR SBR signalling.
void FAAD::AACPlayer::SetDecoderConfig(faacDecHandle& decoder, const unsigned char& defObjectType, const unsigned char& outputFormat, const unsigned char& downMatrix, const unsigned char& useOldADTSFormat, const unsigned char& dontUpSampleImplicitSBR)
{
    faacDecConfigurationPtr decoderConfig = faacDecGetCurrentConfiguration(decoder);
    decoderConfig->defObjectType = defObjectType;
    decoderConfig->outputFormat = outputFormat;
    decoderConfig->downMatrix = downMatrix;
    decoderConfig->useOldADTSFormat = useOldADTSFormat;
    decoderConfig->dontUpSampleImplicitSBR = dontUpSampleImplicitSBR;
    faacDecSetConfiguration(decoder, decoderConfig);
}

/// @brief Callback function used in mp4ff_callback_t.
uint32_t MP4ReadCallback(void* userData, void* buffer, uint32_t length)
{
    return std::fread(buffer, 1, length, static_cast<FILE*>(userData));
}

/// @brief Callback function used in mp4ff_callback_t.
uint32_t MP4SeekCallback(void* userData, uint64_t position)
{
    return std::fseek(static_cast<FILE*>(userData), position, SEEK_SET);
}

/// @brief Create AACPlayer object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateAACPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new FAAD::AACPlayer(mainManager));
}

