// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_aac.cc
/// @brief Class for mpeg 4 tag processing source file.

#include <string>

#include "../include/tag_aac.h"
#include "../include/mainmanager.h"
#include "../include/scopeguard.h"
#include "../include/misc.h"

int AACTag::adtsSampleRate[] = {96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 7350, 0, 0, 0};

AACTag::AACTag(const std::string& filename, const MainManager* const mainManager)
: filename_(filename),
    process_(false),
    mp4File_(false),
    inFile_(NULL),
    mainManager_(mainManager)
{
    inFile_ = std::fopen(filename.c_str(), "rb");

    if(inFile_ != NULL)
	mp4File_ = FAAD::AACPlayer::IsMP4File(inFile_);
}

AACTag::~AACTag()
{
    if(inFile_ != NULL)
	std::fclose(inFile_);
}

/// @brief Get information from AAC type file.
///
/// @todo Need testing.
/// @todo Reduce duplicated codes.
void AACTag::ProcessAAC()
{
    FAAD::aac_buffer aacBuffer;
    aacBuffer.infile = inFile_;

    std::fseek(aacBuffer.infile, 0, SEEK_END);
    int fileRead = std::ftell(aacBuffer.infile);
    std::fseek(aacBuffer.infile, 0, SEEK_SET);

    aacBuffer.buffer = static_cast<unsigned char*>(std::malloc(FAAD_MIN_STREAMSIZE));
    if(!aacBuffer.buffer)
	return;
    std::memset(aacBuffer.buffer, 0, FAAD_MIN_STREAMSIZE);

    int bread = std::fread(aacBuffer.buffer, 1, FAAD_MIN_STREAMSIZE, aacBuffer.infile);
    aacBuffer.bytes_into_buffer = bread;
    aacBuffer.bytes_consumed = 0;
    aacBuffer.file_offset = 0;

    if(bread != FAAD_MIN_STREAMSIZE)
	aacBuffer.at_eof = 1;

    int tagSize = 0;
    if(memcmp(aacBuffer.buffer, "ID3", 3) != 0) {
	tagSize = (aacBuffer.buffer[6] << 21) | (aacBuffer.buffer[7] << 14) |
	    (aacBuffer.buffer[8] << 7) | (aacBuffer.buffer[9] << 0);

	tagSize += 10;
	FAAD::AACPlayer::AACAdvanceBuffer(&aacBuffer, tagSize);
	FAAD::AACPlayer::AACFillBuffer(&aacBuffer);
    }

    float length = 0;
    if(aacBuffer.buffer[0] == 0xFF && (aacBuffer.buffer[1] & 0xF6) == 0xF0) {
	int frames = 0;
	int frameLength = 0;
	int sampleRate = 0;
	float framesPerSec = 0.0f;
	for(;; ++frames) {
	    FAAD::AACPlayer::AACFillBuffer(&aacBuffer);
	    if(aacBuffer.bytes_into_buffer > 7) {
		if(!(aacBuffer.buffer[0] == 0xFF && (aacBuffer.buffer[1] & 0xF6) == 0XF0))
		    break;

		if(frames == 0)
		    sampleRate = adtsSampleRate[(aacBuffer.buffer[2]&0x3c)>>2];

		frameLength = ((static_cast<unsigned int>(aacBuffer.buffer[3]) & 0x3) << 11) |
		    (static_cast<unsigned int>(aacBuffer.buffer[4]) << 3) |
		    aacBuffer.buffer[5] >> 5;

		if(frameLength > aacBuffer.bytes_into_buffer)
		    break;

		FAAD::AACPlayer::AACAdvanceBuffer(&aacBuffer, frameLength);

	    } else
		break;
	}

	framesPerSec = sampleRate / 1024.0f;
	if(framesPerSec > 0.0f)
	    length = frames / framesPerSec;

    } else if(std::memcmp(aacBuffer.buffer, "ADIF", 4) == 0) {
	const int skipSize = (aacBuffer.buffer[4] & 0x80) ? 9 : 0;
	const int bitRate = (static_cast<unsigned int>(aacBuffer.buffer[4 + skipSize] & 0x0F)<<19) |
	    (static_cast<unsigned int>(aacBuffer.buffer[5 + skipSize])<<11) |
	    (static_cast<unsigned int>(aacBuffer.buffer[6 + skipSize])<<3) |
	    (static_cast<unsigned int>(aacBuffer.buffer[7 + skipSize]) & 0xE0);

	length = fileRead;
	if(length != 0.0f && bitRate != 0)
	    length = length * 8.0f / bitRate + 0.5f;
    }

    totalTimeSeconds_ = static_cast<unsigned long>(length);

    title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());

}

void AACTag::ProcessAll()
{
    if(inFile_ != NULL) {
	if(mp4File_) {
	    MP4 mp4(filename_, inFile_, mainManager_);

	    mp4.GetTotalTimeSeconds(totalTimeSeconds_);
	    mp4.GetTitle(title_);
	    mp4.GetArtist(artist_);
	    mp4.GetAlbum(album_);
	    mp4.GetTrack(track_);
	    mp4.GetDate(year_);
	    mp4.GetGenre(genre_);
	    mp4.GetComment(comment_);
	    
	} else
	    ProcessAAC();
    }
    
    process_ = true;
}

void AACTag::ProcessYear()
{
    if(year_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessTrack()
{
    if(track_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessTitle()
{
    if(title_.empty()) {
	if(!process_)
	    ProcessAll();
	else
	    title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
    }
}

void AACTag::ProcessAlbum()
{
    if(album_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessGenre()
{
    if(genre_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessArtist()
{
    if(artist_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessComment()
{
    if(comment_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessTotalTime()
{

    if(totalTime_.empty()) {

	if(totalTimeSeconds_ == 1 && !process_)
	    ProcessAll();

	totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);	

    }
}

AACTag::MP4::MP4(const std::string& filename, FILE* const inFile, const MainManager* const mainManager)
: filename_(filename),
track_(-1),
frameSize_(1024),
valid_(true),
decoder_(faacDecOpen()),
mp4ffFile_(NULL),
mp4Cb_(FAAD::AACPlayer::SetMP4ffCallback(inFile)),
mainManager_(mainManager)
{
    valid_ = Init();
}

AACTag::MP4::~MP4()
{
    if(mp4ffFile_ != NULL)
	mp4ff_close(mp4ffFile_);

    faacDecClose(decoder_);
}

bool AACTag::MP4::Init()
{
    mp4ffFile_ = mp4ff_open_read(mp4Cb_.get());
    if(mp4ffFile_ == NULL)
	return false;

    track_ = FAAD::AACPlayer::GetAACTrack(mp4ffFile_);
    if(track_ < 0)
	return false;

    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;
    mp4ff_get_decoder_config(mp4ffFile_, track_, &buffer, &bufferSize);

    unsigned long sampleRate = 0;
    unsigned char channels = 0;
    if(faacDecInit2(decoder_, buffer, bufferSize, &sampleRate, &channels) < 0)
	return false;

    if(buffer != NULL) {
	if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC_) >= 0) {
	    if(mp4ASC_.frameLengthFlag == 1)
		frameSize_ = 960;
	    if(mp4ASC_.sbr_present_flag == 1)
		frameSize_ *= 2;	
	}

	free(buffer);
	buffer = NULL;
	bufferSize = 0;
    }

    return true;
}

inline void AACTag::MP4::GetTotalTimeSeconds(unsigned long& totalTimeSeconds)
{
    if(valid_) {
	totalTimeSeconds = static_cast<unsigned long>(mp4ff_num_samples(mp4ffFile_, track_) * (frameSize_ - 1.0) / mp4ASC_.samplingFrequency);
    }
}

void AACTag::MP4::GetTitle(Glib::ustring& title)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_title(mp4ffFile_, &value)) {
	    if(value != NULL) {
		title = value;
		free(value);
		value = NULL;

		if(StringIsSpace(title))
		    StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
		else	
		    StringToUTF8(title, Glib::path_get_basename(filename_), mainManager_->GetAssumeCharset());
	    }
	}
    }
}

void AACTag::MP4::GetArtist(Glib::ustring& artist)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_artist(mp4ffFile_, &value)) {
	    if(value != NULL) {
		artist = value;
		free(value);
		value = NULL;

		StringToUTF8(artist, "", mainManager_->GetAssumeCharset());
	    }
	}
    }
}

void AACTag::MP4::GetAlbum(Glib::ustring& album)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_album(mp4ffFile_, &value)) {
	    if(value != NULL) {
		album = value;
		free(value);
		value = NULL;

		StringToUTF8(album, "", mainManager_->GetAssumeCharset());
	    }
	}
    }
}

void AACTag::MP4::GetTrack(Glib::ustring& track)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_track(mp4ffFile_, &value)) {
	    if(value != NULL) {
		track = value;
		free(value);
		value = NULL;

		StringToUTF8(track, "", mainManager_->GetAssumeCharset());
	    }
	}
    }
}

void AACTag::MP4::GetDate(Glib::ustring& date)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_date(mp4ffFile_, &value)) {
	    if(value != NULL) {
		date = value;
		free(value);
		value = NULL;

		StringToUTF8(date, "", mainManager_->GetAssumeCharset());
	    }
	}
    }
}

void AACTag::MP4::GetGenre(Glib::ustring& genre)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_genre(mp4ffFile_, &value)) {
	    if(value != NULL) {
		genre = value;
		free(value);
		value = NULL;

		StringToUTF8(genre, "", mainManager_->GetAssumeCharset());
	    }
	}
    }
}

void AACTag::MP4::GetComment(Glib::ustring& comment)
{
    if(valid_) {
	char* value = NULL;
	if(mp4ff_meta_get_comment(mp4ffFile_, &value)) {
	    if(value != NULL) {
		comment = value;
		free(value);
		value = NULL;

		StringToUTF8(comment, "", mainManager_->GetAssumeCharset());
	    }
	}
    }
}

/// @brief Create AACTag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateAACTag(const std::string& filename, const MainManager* const mainManager)
{
    return std::auto_ptr<Tag>(new AACTag(filename, mainManager));
}

